from __future__ import annotations

import operator
import os
import re
import subprocess
from functools import cached_property, reduce
from pathlib import Path

from yutto.utils.console.logger import Logger
from yutto.utils.funcutils import Singleton


class FFmpegNotFoundError(Exception):
    def __init__(self):
        super().__init__("请配置正确的 FFmpeg 路径")


class FFmpeg(metaclass=Singleton):
    def __init__(self, ffmpeg_path: str = "ffmpeg"):
        try:
            if subprocess.run([ffmpeg_path], capture_output=True).returncode != 1:
                raise FFmpegNotFoundError()
        except FileNotFoundError:
            raise FFmpegNotFoundError()

        self.path = os.path.normpath(ffmpeg_path)

    def exec(self, args: list[str]):
        cmd = [self.path]
        cmd.extend(args)
        Logger.debug(" ".join(cmd))
        # NOTE(aheadlead): FFmpeg 会谜之从 stdin 读取一个字节，这会让调用 yutto 的 shell 脚本踩到坑
        # 这个行为在目前最新的 FFmpeg 6.0 仍然存在
        return subprocess.run(cmd, stdin=subprocess.DEVNULL, capture_output=True)

    @cached_property
    def version(self) -> str:
        output = self.exec(["-version"]).stdout.decode()
        if match_obj := re.match(r"ffmpeg version (?P<version>(\S+)) Copyright", output):
            return match_obj.group("version")
        return "Unknown version"

    @cached_property
    def video_encodecs(self) -> list[str]:
        results: list[str] = []
        output = self.exec(["-codecs"]).stdout.decode()
        for line in output.split("\n"):
            if match_obj := re.match(r"^\s*[D\.]EV[I\.][L\.][S\.] (?P<vcodec>\S+)", line):
                results.append(match_obj.group("vcodec"))
        output = self.exec(["-encoders"]).stdout.decode()
        for line in output.split("\n"):
            if match_obj := re.match(r"^\s*V[F\.][S\.][X\.][B\.][D\.] (?P<encoder>\S+)", line):
                results.append(match_obj.group("encoder"))
        return results

    @cached_property
    def audio_encodecs(self) -> list[str]:
        results: list[str] = []
        output = self.exec(["-codecs"]).stdout.decode()
        for line in output.split("\n"):
            if match_obj := re.match(r"^\s*[D\.]EA[I\.][L\.][S\.] (?P<acodec>\S+)", line):
                results.append(match_obj.group("acodec"))
        output = self.exec(["-encoders"]).stdout.decode()
        for line in output.split("\n"):
            if match_obj := re.match(r"^\s*A[F\.][S\.][X\.][B\.][D\.] (?P<encoder>\S+)", line):
                results.append(match_obj.group("encoder"))
        return results


def concat_commands(commands: list[list[str]]) -> list[str]:
    return reduce(operator.add, commands, [])


class FFmpegInput:
    def __init__(self, path: Path | str, input_id: int, stream_id: int):
        self.path = Path(path)
        self.input_id = input_id
        self.stream_id = stream_id

    def build(self) -> list[str]:
        return ["-i", str(self.path)]

    def __repr__(self):
        return f"FFmpegInput({self.path})"


class FFmpegVideoInput(FFmpegInput): ...


class FFmpegAudioInput(FFmpegInput): ...


class FFmpegOutput:
    def __init__(self, path: Path | str):
        self.path = path
        self.used_inputs: list[FFmpegInput] = []
        self.vcodec: str | None = None
        self.acodec: str | None = None
        self.cover_input: FFmpegVideoInput | None = None
        self.extra_commands: list[str] = []

    def use(self, input: FFmpegInput):
        self.used_inputs.append(input)
        return self

    def set_vcodec(self, codec: str):
        self.vcodec = codec
        return self

    def set_acodec(self, codec: str):
        self.acodec = codec
        return self

    def set_cover(self, cover: FFmpegVideoInput):
        self.cover_input = cover
        return self

    def with_extra_options(self, command: list[str]):
        self.extra_commands.extend(command)
        return self

    def build(self) -> list[str]:
        selected_inputs = concat_commands([["-map", str(input.input_id)] for input in self.used_inputs])
        vcodec = ["-vcodec", self.vcodec] if self.vcodec else []
        acodec = ["-acodec", self.acodec] if self.acodec else []
        # Refer to `-disposition` opiton in https://www.ffmpeg.org/ffmpeg.html#toc-Main-options
        cover_options = (
            [
                f"-c:v:{self.cover_input.stream_id}",
                "copy",
                f"-disposition:v:{self.cover_input.stream_id}",
                "attached_pic",
            ]
            if self.cover_input
            else []
        )
        # Using double dash to make sure that the output file name is not parsed as an option
        # if the output file name starts with a dash
        return selected_inputs + vcodec + acodec + cover_options + self.extra_commands + ["--"] + [str(self.path)]

    def __repr__(self):
        return f"FFmpegOutput({self.path})"


class FFmpegCommandBuilder:
    def __init__(self):
        self.num_inputs = 0
        self.num_video_stream = 0
        self.num_audio_stream = 0
        self.inputs: list[FFmpegInput] = []
        self.outputs: list[FFmpegOutput] = []
        self.extra_commands: list[str] = []

    def add_video_input(self, path: Path | str):
        input = FFmpegVideoInput(path, self.num_inputs, self.num_video_stream)
        self.num_inputs += 1
        self.num_video_stream += 1
        self.inputs.append(input)
        return input

    def add_audio_input(self, path: Path | str):
        input = FFmpegAudioInput(path, self.num_inputs, self.num_audio_stream)
        self.num_inputs += 1
        self.num_audio_stream += 1
        self.inputs.append(input)
        return input

    def with_extra_options(self, command: list[str]):
        self.extra_commands.extend(command)
        return self

    def add_output(self, path: Path | str):
        output = FFmpegOutput(path)
        self.outputs.append(output)
        return output

    def build(self):
        input_commands = concat_commands([input.build() for input in self.inputs])
        output_commands = concat_commands([output.build() for output in self.outputs])
        return input_commands + self.extra_commands + output_commands

    def __repr__(self):
        return "FFmpegCommandBuilder()"



from __future__ import annotations

import argparse
import re
import sys

import httpx

from yutto._typing import EpisodeData, EpisodeId
from yutto.api.bangumi import get_bangumi_list, get_season_id_by_episode_id
from yutto.exceptions import (
    ErrorCode,
    HttpStatusError,
    NoAccessPermissionError,
    NotFoundError,
    UnSupportedTypeError,
)
from yutto.extractor._abc import SingleExtractor
from yutto.extractor.common import extract_bangumi_data
from yutto.utils.asynclib import CoroutineWrapper
from yutto.utils.console.logger import Badge, Logger


class BangumiExtractor(SingleExtractor):
    """番剧单话"""

    REGEX_EP = re.compile(r"https?://www\.bilibili\.com/bangumi/play/ep(?P<episode_id>\d+)")

    REGEX_EP_ID = re.compile(r"ep(?P<episode_id>\d+)")

    episode_id: EpisodeId

    def resolve_shortcut(self, id: str) -> tuple[bool, str]:
        matched = False
        url = id
        if match_obj := self.REGEX_EP_ID.match(id):
            url = f"https://www.bilibili.com/bangumi/play/ep{match_obj.group('episode_id')}"
            matched = True
        return matched, url

    def match(self, url: str) -> bool:
        if match_obj := self.REGEX_EP.match(url):
            self.episode_id = EpisodeId(match_obj.group("episode_id"))
            return True
        else:
            return False

    async def extract(
        self, client: httpx.AsyncClient, args: argparse.Namespace
    ) -> CoroutineWrapper[EpisodeData | None] | None:
        season_id = await get_season_id_by_episode_id(client, self.episode_id)
        bangumi_list = await get_bangumi_list(client, season_id)
        Logger.custom(bangumi_list["title"], Badge("番剧", fore="black", back="cyan"))
        try:
            for bangumi_item in bangumi_list["pages"]:
                if bangumi_item["episode_id"] == self.episode_id:
                    bangumi_list_item = bangumi_item
                    break
            else:
                Logger.error("在列表中未找到该剧集")
                sys.exit(ErrorCode.EPISODE_NOT_FOUND_ERROR.value)

            return CoroutineWrapper(
                extract_bangumi_data(
                    client,
                    bangumi_list_item,
                    args,
                    {
                        "title": bangumi_list["title"],
                    },
                    "{name}",
                )
            )
        except (NoAccessPermissionError, HttpStatusError, UnSupportedTypeError, NotFoundError) as e:
            Logger.error(e.message)
            return None



from __future__ import annotations

import math

from httpx import AsyncClient

from yutto._typing import AvId, BvId, FavouriteMetaData, FId, MId, SeriesId
from yutto.api.user_info import encode_wbi, get_wbi_img
from yutto.exceptions import NotLoginError
from yutto.utils.console.logger import Logger
from yutto.utils.fetcher import Fetcher


# 个人空间·全部
async def get_user_space_all_videos_avids(client: AsyncClient, mid: MId) -> list[AvId]:
    space_videos_api = "https://api.bilibili.com/x/space/wbi/arc/search"
    # ps 随机设置有时会出现错误，因此暂时固定在 30
    # ps: int = random.randint(3, 6) * 10
    ps = 30
    pn = 1
    total = 1
    all_avid: list[AvId] = []
    wbi_img = await get_wbi_img(client)
    while pn <= total:
        params = {
            "mid": mid,
            "ps": ps,
            "tid": 0,
            "pn": pn,
            "order": "pubdate",
        }
        params = encode_wbi(params, wbi_img)
        json_data = await Fetcher.fetch_json(client, space_videos_api, params=params)
        assert json_data is not None
        total = math.ceil(json_data["data"]["page"]["count"] / ps)
        pn += 1
        all_avid += [BvId(video_info["bvid"]) for video_info in json_data["data"]["list"]["vlist"]]
    return all_avid


# 个人空间·用户名
async def get_user_name(client: AsyncClient, mid: MId) -> str:
    wbi_img = await get_wbi_img(client)
    params = {"mid": mid}
    params = encode_wbi(params, wbi_img)
    space_info_api = "https://api.bilibili.com/x/space/wbi/acc/info"
    await Fetcher.touch_url(client, "https://www.bilibili.com")
    user_info = await Fetcher.fetch_json(client, space_info_api, params=params)
    assert user_info is not None
    if user_info["code"] == -404:
        Logger.warning(f"用户 {mid} 不存在，疑似注销或被封禁")
        return f"「用户{mid}」"
    elif user_info["code"] != 0:
        Logger.error(f"获取用户名失败，错误信息：{user_info['message']}，可尝试添加参数 `-c` 登录账号后重试")
    return user_info["data"]["name"]


# 个人空间·收藏夹·信息
async def get_favourite_info(client: AsyncClient, fid: FId) -> FavouriteMetaData:
    api = "https://api.bilibili.com/x/v3/fav/folder/info?media_id={fid}"
    json_data = await Fetcher.fetch_json(client, api.format(fid=fid))
    assert json_data is not None
    data = json_data["data"]
    return FavouriteMetaData(title=data["title"], fid=FId(str(data["id"])))


# 个人空间·收藏夹·avid
async def get_favourite_avids(client: AsyncClient, fid: FId) -> list[AvId]:
    api = "https://api.bilibili.com/x/v3/fav/resource/ids?media_id={fid}"
    json_data = await Fetcher.fetch_json(client, api.format(fid=fid))
    assert json_data is not None
    return [BvId(video_info["bvid"]) for video_info in json_data["data"]]


# 个人空间·收藏夹·全部
async def get_all_favourites(client: AsyncClient, mid: MId) -> list[FavouriteMetaData]:
    api = "https://api.bilibili.com/x/v3/fav/folder/created/list-all?up_mid={mid}"
    json_data = await Fetcher.fetch_json(client, api.format(mid=mid))
    assert json_data is not None
    if not json_data["data"]:
        return []
    return [FavouriteMetaData(title=data["title"], fid=FId(str(data["id"]))) for data in json_data["data"]["list"]]


# 个人空间·视频列表·avid
async def get_medialist_avids(client: AsyncClient, series_id: SeriesId, mid: MId) -> list[AvId]:
    api = "https://api.bilibili.com/x/series/archives?mid={mid}&series_id={series_id}&only_normal=true&pn={pn}&ps={ps}"
    ps = 30
    pn = 1
    total = 1
    all_avid: list[AvId] = []

    while pn <= total:
        url = api.format(series_id=series_id, mid=mid, ps=ps, pn=pn)
        json_data = await Fetcher.fetch_json(client, url)
        assert json_data is not None
        total = math.ceil(json_data["data"]["page"]["total"] / ps)
        pn += 1
        all_avid += [BvId(video_info["bvid"]) for video_info in json_data["data"]["archives"]]
    return all_avid


# 个人空间·视频列表·标题
async def get_medialist_title(client: AsyncClient, series_id: SeriesId) -> str:
    api = "https://api.bilibili.com/x/v1/medialist/info?type=5&biz_id={series_id}"
    json_data = await Fetcher.fetch_json(client, api.format(series_id=series_id))
    assert json_data is not None
    return json_data["data"]["title"]


# 个人空间·稍后再看
async def get_watch_later_avids(client: AsyncClient) -> list[AvId]:
    api = "https://api.bilibili.com/x/v2/history/toview/web"
    json_data = await Fetcher.fetch_json(client, api)
    assert json_data is not None
    if json_data["code"] in [-101, -400]:
        raise NotLoginError("账号未登录，无法获取稍后再看列表哦~ Ծ‸Ծ")
    # TODO: 处理其他code不为0的异常
    return [BvId(video_info["bvid"]) for video_info in json_data["data"]["list"]]



from __future__ import annotations

import sys
from enum import Enum
from types import TracebackType
from typing import Union


class ErrorCode(Enum):
    # 发生错误
    HTTP_STATUS_ERROR = 10
    NO_ACCESS_PERMISSION_ERROR = 11
    UNSUPPORTED_TYPE_ERROR = 12
    WRONG_ARGUMENT_ERROR = 13
    WRONG_URL_ERROR = 14
    EPISODE_NOT_FOUND_ERROR = 15
    MAX_RETRY_ERROR = 16
    NOT_FOUND_ERROR = 17
    NOT_LOGIN_ERROR = 18

    # 异常状况，但并不算错误
    PAUSED_DOWNLOAD = 101


class SuccessCode(Enum):
    SUCCESS = 0


ReturnCode = Union[ErrorCode, SuccessCode]


class YuttoBaseException(Exception):
    code: ErrorCode
    message: str

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class HttpStatusError(YuttoBaseException):
    code = ErrorCode.HTTP_STATUS_ERROR


class NoAccessPermissionError(YuttoBaseException):
    code = ErrorCode.NO_ACCESS_PERMISSION_ERROR


class UnSupportedTypeError(YuttoBaseException):
    code = ErrorCode.UNSUPPORTED_TYPE_ERROR


class MaxRetryError(YuttoBaseException):
    code = ErrorCode.MAX_RETRY_ERROR


class NotFoundError(YuttoBaseException):
    code = ErrorCode.NOT_FOUND_ERROR


class NotLoginError(YuttoBaseException):
    code = ErrorCode.NOT_LOGIN_ERROR


def handleUncaughtException(exctype: type[Exception], exception: Exception, trace: TracebackType):
    oldHook(exctype, exception, trace)
    if isinstance(exception, YuttoBaseException):
        sys.exit(exception.code.value)


sys.excepthook, oldHook = handleUncaughtException, sys.excepthook


if __name__ == "__main__":
    try:
        raise HttpStatusError("HTTP 错误")
    except (HttpStatusError, UnSupportedTypeError) as e:
        print(e.code.value, e.message)
        raise e



from __future__ import annotations

from typing import Any


class Undefined: ...


def data_has_chained_keys(data: Any, keys: list[str]) -> bool:
    if isinstance(data, Undefined):
        return False
    if not keys:
        return True
    if not isinstance(data, dict):
        return False
    key, *remaining_keys = keys
    return data_has_chained_keys(data.get(key, Undefined()), remaining_keys)  # type: ignore
